
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "lut.h"

/*==============================================================*/
/* load a linear lut into the struct                            */
void lut_init(lut_t *lut)
{
	int a;

	lut->gamma = 1.;

	for (a=0; a<256; a++)
		lut->lut8[a] = a;

	for (a=0; a<65536; a++)
		lut->lut16[a] = a;
}

/*===============================================================*/
/* adjust the value in the lut using new=old^(1./gamma)          */
void lut_gamma(lut_t *lut, float gam)
{
	int a;
	float f, power;

	if (gam < .01)
	{
		fprintf(stderr, "gamma of  <= 0 is not valid\n");
		return;
	}

	lut->gamma = gam;

	power = 1./gam;
	for (a=0; a<256; a++)
	{
		f = (float)lut->lut8[a] / 255.;
		f = pow(f, power);
		if (f > 1.) f = 1;
		if (f < 0.) f = 0; /* sanity check */
		lut->lut8[a] = 255.*f;
	}

	for (a=0; a<65536; a++)
	{
		f = (float)lut->lut16[a] / 65535.;
		f = pow(f, power);
		if (f > 1.) f = 1;
		if (f < 0.) f = 0; /* sanity check */
		lut->lut16[a] = 65535.*f;
	}
}

