/* main.c 
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <X11/Xlib.h> 
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/extensions/XShm.h>

#include <glib.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sched.h>
#include <pthread.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "timer.h"
#include "platform.h"
#include "lut.h"
#include "images.h"
#include "opts.h"

#include "read_ppm.h"
#include "read_rla.h"
#include "read_mayaiff.h"
#include "read_cin.h"
#include "read_sgi.h"
#ifdef HAVE_LIBTIFF
#include "read_tiff.h"
#endif

#define SCROLL_BAR_HEIGHT 	10

/*  keysym values.. at least on my box */
#define SPACE_BAR			32
#define F_KEY				102
#define Q_KEY				113
#define H_KEY				104
#define C_KEY				99
#define A_KEY				97
#define R_KEY				114
#define G_KEY				103
#define B_KEY				98 
#define U_KEY				117
#define PERIOD_KEY			46	
#define COMMA_KEY			44
#define SLASH_KEY			47	
#define ESC					65307
#define UP_ARROW			65362
#define UP_KP				65431
#define DOWN_ARROW			65364
#define DOWN_KP				65433
#define LEFT_ARROW			65361
#define LEFT_KP				65430
#define RIGHT_ARROW			65363
#define RIGHT_KP			65432

#define LOADING				1
#define PLAYING				2
#define STOPPED				3

#define DISP_RGB			0
#define DISP_RED			1
#define DISP_GREEN			2
#define DISP_BLUE			3
#define DISP_ALPHA			4

Display *display;
Window rootw, win, sb_win, st_win, main_win;
GC gc, sb_gc, st_gc;	
XSizeHints whisper;

int gargc;
char **gargv;

XImage *img;
XShmSegmentInfo shminfo;
XImage *this_img, **img_list;
XShmSegmentInfo *shminfo_list;

XFontStruct *fnt;
XColor sb_back, sb_fore;
XColor sb_drop_hi, sb_drop_lo;
XColor sb_stale_hi, sb_stale_lo;
XVisualInfo *pgob, gob;

int width;
int height;
int text_height;

int gnfiles = 0;

int status = 0;
int hidden = 0;

/* misc globals to control which way that we playback, and
 * if we should stop once we hit the end */
int play_direction = 1;
int stop_at_end = 0;

double frames_a_second  = 24.00;
double time_for_a_frame = 1./24.00;

int start_bound = -1;
int end_bound   = 99999999;

int disp_what = 0;
lut_t load_lut;

/* a map of who was dropped the last time around */
char *dropped_nums;

/* a map of who is stale, and the timestamps */
char *stale_files;
time_t *timestamps = NULL;

/* if the timestamp thread should reload files when
 * it finds stale fish  */
int auto_reload_stale = 0;
int auto_check_stale = 0;

unsigned char **image_list; /* depreciated */

void * usages[] = 
{
#ifdef HAVE_LIBTIFF
	UsageTIFF,
#endif
	UsagePPM,
	UsageMIFF,
	UsageRLA,
	UsageCinDpx,
	UsageSGI
};
int nusages = sizeof(usages)/sizeof(void *);

int timestamp_thread_live = 1;

/*==============================================
 * cleanup memory allocated for shm images     */
void cleanup()
{
	int a;

	/* cleanup! */
	for (a=0; a<gnfiles; a++)
	{
		if (img_list[a] == NULL) continue;

		XShmDetach(display, &shminfo_list[a]);
		XDestroyImage(img_list[a]);
		shmdt(shminfo_list[a].shmaddr);
		shmctl(shminfo_list[a].shmid, IPC_RMID, 0);
	}
}

/*==============================================
 * handler for SIGINT       
 *
 * this is only installed after img_list is
 * alloced, so we shouldnt segfault, or have to check */
void term_handler(int b)
{
	/* to prevent multiple getting multiple sigs and puking */
	signal(SIGINT, SIG_IGN);

	cleanup();

	exit(1);
}

/*==============================================*/
void SetST(int curr_frame, int nframes)
{
	static int extent = -1;
	char str1[30], str2[30];
	int tmp;

	tmp = curr_frame / frames_a_second;
	sprintf(str1, "%05d/%05d  %02d:%02d.%02d", 
				curr_frame+1, nframes, 
				tmp/60, tmp%60, 
				curr_frame%(int)(frames_a_second));

	if (extent == -1)
	{
		extent = XTextWidth(fnt, str1, strlen(str1));
	}

	if (status == LOADING)
		sprintf(str2, "%d x %d DISK %.2f fps", width, height, frames_a_second);
	else
		sprintf(str2, "%d x %d RAM  %.2f fps", width, height, frames_a_second);

	XClearWindow(display, st_win);
	XDrawString(display, st_win, st_gc, width-10-extent, text_height/2+fnt->ascent, str1, strlen(str1));
	XDrawString(display, st_win, st_gc, 10, text_height/2+fnt->ascent, str2, strlen(str2));
}

/*==============================================*/
void SetSBFrame(int frame, int nframes)
{
	static int first = 1;
	static int draw_width = 0;
	static int *draw_starts = NULL;
	int a;

	if (first)
	{
		first = 0;
		draw_width = ceil(1./(double)(nframes)*width);
		draw_starts = (int *)malloc(nframes*sizeof(int));
		for (a=0; a<nframes; a++)
			draw_starts[a] = floor((double)a/(double)(nframes)*width);
	}

	XFillRectangle(display, sb_win, sb_gc, draw_starts[frame], 0, 
							draw_width, SCROLL_BAR_HEIGHT);

/*
	XFillRectangle(display, sb_win, sb_gc, floor((double)frame/(double)(nframes)*width), 0, 
							ceil(1./(double)(nframes)*width), SCROLL_BAR_HEIGHT);
*/
}

/*==============================================*/
void SetSB(int curr_frame, int nframes, int do_others)
{
	int a, switched;

	if (hidden) return;

	XClearWindow(display, sb_win);

	if (do_others)
	{
		/* draw all the dropped */
		XSetForeground(display, sb_gc, sb_drop_lo.pixel);
		for (a=0; a<nframes; a++)
		{
			if (a == curr_frame) continue;
	
			if (dropped_nums[a])
				SetSBFrame(a, nframes);
		}

		/* draw all the stale */
		XSetForeground(display, sb_gc, sb_stale_lo.pixel);
		for (a=0; a<nframes; a++)
		{
			if (a == curr_frame) continue;

			if (stale_files[a])
				SetSBFrame(a, nframes);
		}
	}


	/* draw the current */
	if (stale_files[curr_frame])
		XSetForeground(display, sb_gc, sb_stale_hi.pixel);
	else
		XSetForeground(display, sb_gc, sb_fore.pixel);

	/* XFillRectangle(display, sb_win, sb_gc, 0, 0, (double)curr_frame/(double)(nframes-1)*width, SCROLL_BAR_HEIGHT); */
	SetSBFrame(curr_frame, nframes);

	SetST(curr_frame, nframes);

}

/*==============================================
 * returns the number of stale files found      */
int TimestampCheck(char **file_list, int nfiles)
{
	int a, retval;
	struct stat stat_buf;

	retval = 0;

	for (a=0; a<nfiles; a++)
	{
		stat(file_list[a], &stat_buf);
		if (stat_buf.st_ctime != timestamps[a])
		{
			stale_files[a] = 1;		
			retval++;
		}
	}	

	return retval;
}

/*==============================================
 * reload all the stale frames...               */
void ReloadStale(char **file_list, int nfiles)
{
	int a, b, pos;
	struct stat stat_buf;

	if (!TimestampCheck(file_list, nfiles)) return;

	for (a=0; a<nfiles; a++)
	{
		if (!stale_files[a]) continue;

		stale_files[a] = 0;
		/* get the timestamp for the file right before loading it */
		stat(file_list[a], &stat_buf);
		memcpy(&timestamps[a], &stat_buf.st_ctime, sizeof(time_t));

		/* get the new image from the loader */
		if (!ReadFiles(file_list, a, &width, &height, image_list, gargc, gargv, &load_lut))
		{
			fprintf(stderr, "error reading %s\n", file_list[a]);

			/* if the frame is bad, stick in a blue screen of death 
			 * free here because the loader may not have allocted
			 * the buffer to be the correct size */
			if (image_list[a])
				free(image_list[a]);

			pos = 4*width*height;
			image_list[a] = (unsigned char *)malloc(pos);
			memset(image_list[a], 0, pos);
			if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
			{
				for (b=0; b<pos; b+=4)
					image_list[a][b] = 255;
			}
			else
			{
				for (b=1; b<pos; b+=4)
					image_list[a][b] = 255;
			}
		}

		/* and stick it into the ximage */
		memcpy(img_list[a]->data, image_list[a], img_list[a]->bytes_per_line*img_list[a]->height);

		free(image_list[a]);
	}

}

/*==============================================*/
void * TimestampThreadMain(void *arg)
{
	int has_stale;
	char **file_list = (char **)arg;
	int nfiles = timestamp_thread_live;

	while (timestamp_thread_live)
	{
		has_stale = TimestampCheck(file_list, nfiles);	
	
		if ((has_stale) && (auto_reload_stale))
		{
			ReloadStale(file_list, nfiles);
		}

		sleep(5);
	}

	pthread_exit(NULL);
}

/*==============================================*/
void GoX(int argc, char **argv)
{
	XEvent report;
	unsigned long eventmask;
	int screen;
	XGCValues values;
	XSetWindowAttributes attrib;
	XColor white, black;
	int bsz, count;
	unsigned long valuemask;
	char title[30];

   /* 
    open(NULL) means get display name from environment 
    variable DISPLAY, set by xterm 
   */
	if((display=XOpenDisplay(NULL))==NULL)
	{
		fprintf(stderr,"can't open\n");
		exit(1);
	}

	/* check for shm extention */
	if (!XShmQueryExtension(display))
	{
		fprintf(stderr, "XShm extention not found!\n");
		exit(1);
	}

	fnt = XLoadQueryFont(display, "fixed");
	text_height = fnt->descent + fnt->ascent;

	screen = DefaultScreen(display);
	rootw = RootWindow(display,screen);

	gob.class = TrueColor;
	gob.depth = 24;
	valuemask = 0;
	valuemask |= VisualClassMask;
	valuemask |= VisualDepthMask;

	pgob = XGetVisualInfo(display, valuemask, &gob, &count);
	if (!count)
	{
		fprintf(stderr, "24bpp truecolor visuals not found!\n");
		return;
	}

	attrib.colormap = XCreateColormap(display, rootw, pgob->visual,
                                  AllocNone);

	white.red = 0xffff;
	white.green = 0xffff;
	white.blue = 0xffff;
	black.red = 0;
	black.green = 0;
	black.blue = 0;
	sb_back.red   = 0x6400;
	sb_back.green = 0x6900;
	sb_back.blue  = 0x8400;
	sb_fore.red   = 0x8100;
	sb_fore.green = 0x6200;
	sb_fore.blue  = 0x6200;
	sb_drop_lo.red   = 0x6600;
	sb_drop_lo.green = 0x8200;
	sb_drop_lo.blue  = 0x6200;
	sb_drop_hi.red   = 0x8f00;
	sb_drop_hi.green = 0xb500;
	sb_drop_hi.blue  = 0x8900;
	sb_stale_lo.red    = 0x9e00;
	sb_stale_lo.green  = 0x3400;
	sb_stale_lo.blue   = 0x3400;
	sb_stale_hi.red    = 0xce00;
	sb_stale_hi.green  = 0x4400;
	sb_stale_hi.blue   = 0x4400;

	XAllocColor(display, attrib.colormap, &white);
	XAllocColor(display, attrib.colormap, &black);
	XAllocColor(display, attrib.colormap, &sb_back);
	XAllocColor(display, attrib.colormap, &sb_fore);
	XAllocColor(display, attrib.colormap, &sb_drop_hi);
	XAllocColor(display, attrib.colormap, &sb_drop_lo);
	XAllocColor(display, attrib.colormap, &sb_stale_hi);
	XAllocColor(display, attrib.colormap, &sb_stale_lo);

	attrib.background_pixel = black.pixel;
	attrib.border_pixel = black.pixel;

	bsz=0;

	main_win = XCreateWindow(display, rootw, 0, 0, width, height+SCROLL_BAR_HEIGHT+2*text_height, bsz, pgob->depth,
                    InputOutput, pgob->visual, 
                    CWBorderPixel|CWColormap|CWBackPixel, &attrib);

	win = XCreateWindow(display, main_win, 0, 0, width, height, bsz, pgob->depth,
                    InputOutput, pgob->visual, 
                    CWBorderPixel|CWColormap|CWBackPixel, &attrib);

	attrib.background_pixel = sb_back.pixel;

	sb_win = XCreateWindow(display, main_win, 0, height, width, SCROLL_BAR_HEIGHT, bsz, pgob->depth,
                    InputOutput, pgob->visual, 
                    CWBorderPixel|CWColormap|CWBackPixel, &attrib);

	attrib.background_pixel = black.pixel;

	st_win = XCreateWindow(display, main_win, 0, height+SCROLL_BAR_HEIGHT, width, 2*text_height, bsz, pgob->depth,
                    InputOutput, pgob->visual, 
                    CWBorderPixel|CWColormap|CWBackPixel, &attrib);

	whisper.flags = PSize | PPosition | PMinSize | PMaxSize | PResizeInc;
	whisper.width = width;
	whisper.height = height+SCROLL_BAR_HEIGHT+2*text_height;
	whisper.min_width = width;
	whisper.min_height = height+SCROLL_BAR_HEIGHT+2*text_height;
	whisper.max_width = width;
	whisper.max_height = height+SCROLL_BAR_HEIGHT+2*text_height;
	whisper.width_inc = 0;
	whisper.height_inc = 0;

	sprintf(title, "mFlip  %s\n", VERSION);
	XSetStandardProperties(display, main_win, title, "", None, 
                       argv, argc, &whisper);

	gc = XCreateGC(display, win, 0, &values);
	sb_gc = XCreateGC(display, sb_win, 0, &values);
	st_gc = XCreateGC(display, st_win, 0, &values);

	eventmask = ExposureMask | KeyPressMask;
	XSelectInput(display, win, eventmask);

	eventmask = ExposureMask | KeyPressMask | ButtonPressMask | ButtonReleaseMask |
					PointerMotionMask;
	XSelectInput(display, sb_win, eventmask);

	XMapWindow(display, main_win);
	XMapWindow(display, win);
	XMapWindow(display, sb_win);
	XMapWindow(display, st_win);

	XNextEvent(display,&report);
	for (count=0; count<4; count++)
	{
		while(report.type!=Expose) 
		{
	      	XNextEvent(display,&report);
		}
	}
	
	XSetForeground(display, sb_gc, sb_fore.pixel);
	XSetForeground(display, st_gc, white.pixel);

	/* setup a shm ximage */
	img = XShmCreateImage(display, pgob->visual, 24,  ZPixmap, NULL, &shminfo, width, height);

	/* do the shm incantation */
	shminfo.shmid = shmget(IPC_PRIVATE, img->bytes_per_line*img->height, IPC_CREAT|0777);

	/* do some attatching */
	shminfo.shmaddr = img->data = shmat(shminfo.shmid, 0, 0);
	shminfo.readOnly = False;
	XShmAttach(display, &shminfo);
}

/*==============================================*/
void usage(char **argv)
{
	int a;
	void (*reader_usage)();

	/* basic options */ 
	printf("usage: %s [params] pattern\n\n", argv[0]);
	printf("\tTiming Parameters:\n");
	printf("\t  -b x\t\tbenchmark mode, 4:3, x by .75x\n");
	printf("\t  -f fps\tspecify the frame rate, fps (default is 24.00)\n");
	printf("\t  -n \t\tNTSC frame rate 29.97\n");
	printf("\t  -p \t\tPAL  frame rate 25.00\n");
	printf("\n\tPattern:\n");
	printf("\t     use # (\\# in some shells) to indicate a numeric sequence\n");
	printf("\t     ie: foo.\\#.tif -> foo.1.tif, foo.2.tif, foo.3.tif, ...\n");
	printf("\t     but it'll also find foo.0001.tif, foo.0002.tif, foo.0003.tif, ...\n");
	printf("\t  -s num\tonly match # >= num (default is lowest found)\n");
	printf("\t  -e num\tonly match # <= num (default is highest found)\n");
	printf("\t  -a num\tset auto-reloading/auto-detection of stale frames:\n");
	printf("\t\t\t\t0 = auto-detect, no auto-reload\n");
	printf("\t\t\t\t1 = auto-detect, auto-reload\n");
	printf("\t\t\t\t  both are disabled by default\n");

	printf("\n\tDisplay related:\n");
	printf("\t  -x scale\tscale x by scale\n");
	printf("\t  -y scale\tscale y by scale\n");
	printf("\t  -S scale\tscale both (hand optimized for 0.25, 0.33, 0.5)\n");
	printf("\t  -g n\t\tset gamma to n\n");
 
	/* extensions by readers */
	for (a=0; a<nusages; a++)
	{
		reader_usage = usages[a];
		(reader_usage)();
	}

	/* interactive stuff */
	printf("\n\tKeys:\n");
	printf("\t  u  \t\tforce reloading of stale frames\n");
	printf("\t  h  \t\thide/unhide the progress bar and frame counter\n");
	printf("\t  f  \t\tdisplay current frame file name on stdout when paused\n");
	printf("\t  q esc\t\tquit\n");
	printf("\t  r  \t\tview only red channel\n");
	printf("\t  g  \t\tview only green channel\n");
	printf("\t  b  \t\tview only blue channel\n");
	printf("\t  a  \t\tview only alpha channel\n");
	printf("\t  c  \t\tview RGB channels\n");
	printf("\t  space\t\tpause/play\n");
	printf("\t  left arrow\trewind 1 frame when paused\n");
	printf("\t  right arrow\tadvance 1 frame when paused\n");
	printf("\t  up arrow\trewind to beginning when paused\n");
	printf("\t  down arrow\tadvance to end when paused\n");
	printf("\t  .\t\tset play direction to forward\n");
	printf("\t  ,\t\tset play direction to backward\n");
	printf("\t  /\t\ttoggle stop/loop at end of playback (default is loop)\n");
	printf("\n");
}

/*==============================================*/
void switch_disp(XImage **img_list, int frameno, int truth)
{
	int a, b, pos;

	dropped_nums[frameno] = 0;

	switch(disp_what)
	{
		case DISP_RGB:
			XShmPutImage(display, win, gc, img_list[frameno], 0, 0, 0, 0, width, height, truth);
			break;

		case DISP_RED:	
			pos = 0;
			if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b], img_list[frameno]->data[pos+b+2], 3);
					pos += img->bytes_per_line;
				}
			}
			else
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b+1], img_list[frameno]->data[pos+b+3], 3);
					pos += img->bytes_per_line;
				}
			}
			XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, truth);
			break;

		case DISP_GREEN:	
			pos = 0;
			if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b], img_list[frameno]->data[pos+b+1], 3);
					pos += img->bytes_per_line;
				}
			}
			else
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b+1], img_list[frameno]->data[pos+b+2], 3);
					pos += img->bytes_per_line;
				}
			}
			XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, truth);
			break;

		case DISP_BLUE:	
			pos = 0;
			if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b], img_list[frameno]->data[pos+b], 3);
					pos += img->bytes_per_line;
				}
			}
			else
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b+1], img_list[frameno]->data[pos+b+1], 3);
					pos += img->bytes_per_line;
				}
			}
			XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, truth);
			break;

		case DISP_ALPHA:	
			pos = 0;
			if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b], img_list[frameno]->data[pos+b+3], 3);
					pos += img->bytes_per_line;
				}
			}
			else
			{
				for (a=0; a<height; a++)
				{
					for (b=0; b<img->bytes_per_line; b+=4)
						memset(&img->data[pos+b+1], img_list[frameno]->data[pos+b], 3);
					pos += img->bytes_per_line;
				}
			}
			XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, truth);
			break;

	}


}
/*==============================================*/
/* try and hack together 'fast mode'.. alloc a 
 * ximage for each frame. keep the old ximage (img)
 * around incase we have to switch to slow mode  */
void main_fast(int argc, char **argv, char **file_list, 
				int nfiles, int bench) 
{
	int a, b, pos;
	int CompletionType, keyhit, sleep_ret;
	int dropped, nframes, count, frameno;
	int scrub = 0;
	double times;
	struct stat stat_buf;
	my_timer timer1, timer2, clocks_for_a_frame;
	XEvent report;
	unsigned long tst_args[2];
	pthread_t tst;
	pthread_attr_t ta;
	struct sched_param parm;

#ifdef X86_LINUX_TIMER
	clocks_for_a_frame = SecsToClocks(time_for_a_frame);
#else
	clocks_for_a_frame.tv_sec  = time_for_a_frame;
	clocks_for_a_frame.tv_usec = time_for_a_frame * 1e6;
#endif

	img_list = (XImage **)malloc(nfiles*sizeof(XImage *));
	shminfo_list = (XShmSegmentInfo *)malloc(nfiles*sizeof(XShmSegmentInfo));

	for (a=0; a<nfiles; a++)
		img_list[a] = NULL;

	gnfiles = nfiles;
	signal(SIGINT, term_handler);

	/* allocation */
	status = LOADING;

	/* allocate some space to hold various things, like a record of all 
	 * frames that were dropped while playing.. should probably move 
	 * this to its own function when we get to loading in new frames  */
	dropped_nums = (char *)malloc(nfiles*sizeof(char));
	memset(dropped_nums, 0, nfiles*sizeof(char));
	timestamps = (time_t *)malloc(nfiles*sizeof(time_t));
	stale_files = (char *)malloc(nfiles*sizeof(char));
	memset(stale_files, 0, nfiles*sizeof(char));

	for (a=0; a<nfiles; a++)
	{

		if (bench)
		{
			if (!a)
				GoX(argc, argv);	
		}
		else
		{
			if (!a)
				image_list = (unsigned char **)malloc(nfiles*sizeof(unsigned char *));

			image_list[a] = NULL;

			/* get the timestamp for the file right before loading it */
			stat(file_list[a], &stat_buf);
			memcpy(&timestamps[a], &stat_buf.st_ctime, sizeof(time_t));

			/* this is an ugly hack. change the loader interface if we like fast mode */
			if (!ReadFiles(file_list, a, &width, &height, image_list, argc, argv, &load_lut))
			{
				fprintf(stderr, "error reading %s\n", file_list[a]);

				/* if the frame is bad, stick in a blue screen of death */
				if (image_list[a])
					free(image_list[a]);
/*
				if (!a)
				{
					width *= scale_x;
					height *= scale_y;
				}
*/

				pos = 4*width*height;
				image_list[a] = (unsigned char *)malloc(pos);
				memset(image_list[a], 0, pos);
				if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
				{
					for (b=0; b<pos; b+=4)
						image_list[a][b] = 255;
				}
				else
				{
					for (b=1; b<pos; b+=4)
						image_list[a][b] = 255;
				}
	
//				exit(1);
			}

			if (!a)
				GoX(argc, argv);

		}


		/* setup a shm ximage */
		img_list[a] = XShmCreateImage(display, pgob->visual, 24,  ZPixmap, NULL, &shminfo_list[a], width, height);
		this_img = img_list[a];

		/* do the shm incantation */
		shminfo_list[a].shmid = shmget(IPC_PRIVATE, this_img->bytes_per_line*this_img->height, IPC_CREAT|0777);

		/* do some attatching */
		shminfo_list[a].shmaddr = this_img->data = shmat(shminfo_list[a].shmid, 0, 0);
		shminfo_list[a].readOnly = False;
		XShmAttach(display, &shminfo_list[a]);

		CompletionType = XShmGetEventBase(display) + ShmCompletion;

		if (bench)
		{
			for (b=0; b<this_img->bytes_per_line*this_img->height; b++)
				this_img->data[b] = drand48()*255.;
		}
		else
		{
			memcpy(this_img->data, image_list[a], this_img->bytes_per_line*this_img->height);
			free(image_list[a]);
		}
		XShmPutImage(display, win, gc, this_img, 0, 0, 0, 0, width, height, True);

		XNextEvent(display, &report);
		while (report.type != CompletionType) 
		{
			switch(report.type)
			{
				case KeyPress:
					keyhit = XKeycodeToKeysym(display, report.xkey.keycode,  0);
					switch(keyhit)
					{
						case Q_KEY:	
						case ESC:
							cleanup();
							exit(1);
					}
					break;
			}
			XNextEvent(display, &report);
		}
		SetSB(a, nfiles, 0);
	}

	status = PLAYING;

	dropped = 0;
	nframes = 200;
	count = -1;
	frameno = 0;
	TimerStart(&timer2);
	TimerStart(&timer1);

	SetSB(frameno, nfiles, 1);
	XShmPutImage(display, win, gc, img_list[frameno], 0, 0, 0, 0, width, height, True);

	/* start up the timestamp checking thread */
	timestamp_thread_live = nfiles;
	if ((!bench) && (auto_check_stale))
	{
		pthread_attr_init(&ta);
		sched_getparam(getpid() , &parm);
		parm.sched_priority = sched_get_priority_min(SCHED_OTHER);
		pthread_attr_setschedparam(&ta, &parm);
		pthread_create(&tst, &ta, TimestampThreadMain, file_list);
	}

	parm.sched_priority = sched_get_priority_max(SCHED_OTHER);
	pthread_setschedparam(pthread_self(), SCHED_OTHER, &parm);
	/* main event loop */
	while (1)
	{
		XNextEvent(display, &report);

		switch(report.type)
		{
			case Expose:
				if (report.xany.window == sb_win)
					SetSB(frameno, nfiles, 1);
				else 
				if (report.xany.window == win)
				{
					/* if we're playing, dont worry about it. we'll fix it next frame 	
					 * otherwise, redisplay this frame but dont generate a new event */	
					if (status != PLAYING)
					{
						switch_disp(img_list, frameno, False);
					}
//						XShmPutImage(display, win, gc, img_list[frameno], 0, 0, 0, 0, width, height, False);
				}
				break;

			/* this only occurs on the SB for now */
			case ButtonPress:
				if (report.xbutton.button != 1) break;
				frameno = nfiles * ((float)report.xbutton.x / (float)width);
				if (frameno >= nfiles) frameno = nfiles-1;
				if (frameno < 0) frameno = 0;
				SetSB(frameno, nfiles, 1);
				switch_disp(img_list, frameno, False);
				if (status == PLAYING)
					scrub = 2;
				else
					scrub = 1;	
				status = STOPPED;
				break;
	
			/* this only occurs on the SB for now */
			case ButtonRelease:
				if (report.xbutton.button != 1) break;
				if (scrub == 2)
				{
					status = PLAYING;
					if (disp_what == DISP_RGB)
						XShmPutImage(display, win, gc, img_list[frameno], 0, 0, 0, 0, width, height, True);
					else
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, True);

					TimerStart(&timer2);
					TimerStart(&timer1);
					count = 0;
				}
				scrub = 0;
				break;
		
			case MotionNotify:
				if (!scrub) break;
				frameno = nfiles * ((float)report.xmotion.x / (float)width);

				if (frameno >= nfiles) frameno = nfiles-1;
				if (frameno < 0) frameno = 0;
				SetSB(frameno, nfiles, 1);
				switch_disp(img_list, frameno, False);
				break;
			case KeyPress:
				keyhit = XKeycodeToKeysym(display, report.xkey.keycode,  0);
				switch(keyhit)
				{
					case SLASH_KEY:
						stop_at_end = !stop_at_end;
						break;

					/* > makes us play forward [keycode for . and > are the same */
					case PERIOD_KEY:
						play_direction = 1;
						break;

					/* < makes up play backwards */
					case COMMA_KEY:
						play_direction = -1;
						break;

					/* U forces a reload of all stale frames */
					case U_KEY:
						ReloadStale(file_list, nfiles);
						break;

					case C_KEY:
						disp_what = DISP_RGB;
						if (status == STOPPED)
							XShmPutImage(display, win, gc, img_list[frameno], 0, 0, 0, 0, width, height, False);
						break;
				
					case A_KEY:
						disp_what = DISP_ALPHA;
						if (status != STOPPED) break;
						switch_disp(img_list, frameno, False);
						break;

					case R_KEY:
						disp_what = DISP_RED;
						if (status != STOPPED) break;
						switch_disp(img_list, frameno, False);
						break;

					case G_KEY:
						disp_what = DISP_GREEN;
						if (status != STOPPED) break;
						switch_disp(img_list, frameno, False);
						break;

					case B_KEY:
						disp_what = DISP_BLUE;
						if (status != STOPPED) break;
						switch_disp(img_list, frameno, False);
						break;


					/* h hides the status bar and info stuff */
					case H_KEY:
						if (!hidden)
						{
							XResizeWindow(display, main_win, width, height);
							whisper.min_width = width;
							whisper.min_height = height;
							whisper.max_width = width;
							whisper.max_height = height;
							XSetStandardProperties(display, main_win, "mFlip", "", None, 
                       				argv, argc, &whisper);
						}
						else
						{
							XResizeWindow(display, main_win, width, height+SCROLL_BAR_HEIGHT+2*text_height);
							whisper.min_width = width;
							whisper.min_height = height+SCROLL_BAR_HEIGHT+2*text_height;
							whisper.max_width = width;
							whisper.max_height = height+SCROLL_BAR_HEIGHT+2*text_height;
							XSetStandardProperties(display, main_win, "mFlip", "", None, 
                       				argv, argc, &whisper);

						}				
						hidden = !hidden;
						break;

					/* space bar starts and stops us */
					case SPACE_BAR:
						if (status == PLAYING)
						{
							status = STOPPED;
						}
						else
						if (status == STOPPED)
						{
							status = PLAYING;
							if (disp_what == DISP_RGB)
								XShmPutImage(display, win, gc, img_list[frameno], 0, 0, 0, 0, width, height, True);
							else
								XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, True);

							TimerStart(&timer2);
							TimerStart(&timer1);
							count = 0;
						}
					break;
		
					/* up takes up  back to the beginning */
					case UP_ARROW:
					case UP_KP:
						if (status != STOPPED) break;
						frameno = 0;
						switch_disp(img_list, frameno, False);

						SetSB(frameno, nfiles, 1);
						break;

					/* down goes to the end */
					case DOWN_ARROW:
					case DOWN_KP:
						if (status != STOPPED) break;
						frameno = nfiles-1;
						switch_disp(img_list, frameno, False);
						SetSB(frameno, nfiles, 1);
						break;

					/* left goes back a frame */
					case LEFT_ARROW:
					case LEFT_KP:
						if (status != STOPPED) break;
						if (frameno == 0) 
							frameno = nfiles-1;
						else
							frameno--;
						switch_disp(img_list, frameno, False);

						SetSB(frameno, nfiles, 1);
						break;

					/* right goes back a frame */
					case RIGHT_ARROW:
					case RIGHT_KP:
						if (status != STOPPED) break;
						frameno = (frameno+1) % nfiles;
						switch_disp(img_list, frameno, False);

						SetSB(frameno, nfiles, 1);
						break;

					/* f displays the current frame name on stdout */
					case F_KEY:
						if (status != STOPPED) break;
						fprintf(stdout, "%s\n", file_list[frameno]);
						break;

					/* TODO: CLEANUP! */
					case ESC:
					case Q_KEY:	
						timestamp_thread_live = 0;
						/* wakeup the silly thing if its sleeping */
						if (auto_check_stale)
						{
							pthread_cancel(tst);
							pthread_join(tst, NULL);
						}
		
						for (a=0; a<nfiles; a++)
						{
							XShmDetach(display, &shminfo_list[a]);
							XDestroyImage(img_list[a]);
							shmdt(shminfo_list[a].shmaddr);
							shmctl(shminfo_list[a].shmid, IPC_RMID, 0);
						}

						XShmDetach(display, &shminfo);
						XDestroyImage(img);
						shmdt(shminfo.shmaddr);
						shmctl(shminfo.shmid, IPC_RMID, 0);

						XDestroySubwindows(display, main_win);
						XDestroyWindow(display, main_win);
						XCloseDisplay(display);
						exit(1);
				}
				break;

			default:
				if (report.type == CompletionType)
				{	
					if (!bench)
					{
						/* this returns the number of frames we dropped, if any */
						sleep_ret = TimerSleepAndReset(&timer1, clocks_for_a_frame);
						if (sleep_ret)
						{

							if (play_direction > 0)
							{
								for (a=0; a<sleep_ret; a++)
								{
									dropped_nums[(frameno+a) % nfiles] = 1;	
								}

							
								count += sleep_ret;
								frameno = (frameno + sleep_ret) % nfiles;
								dropped += sleep_ret;
							}
							else
							{
								for (a=0; a<sleep_ret; a++)
								{
									dropped_nums[(frameno-a) % nfiles] = 1;	
								}

							
								count -= sleep_ret;
								frameno = (frameno - sleep_ret) % nfiles;
								dropped -= sleep_ret;
							}
						}
					}


					if (status == PLAYING)
					{
						/* on LE, the ximage is BGRP 
						 * 	  BE is             PBGR */
						switch_disp(img_list, frameno, True);
						count++;
						//frameno++;
						frameno += play_direction;
						if (frameno >= nfiles) 
						{

							if (stop_at_end) 
							{
								status = STOPPED;
								frameno = nfiles-1;
							}
							else
								frameno = 0;
						}
						else
						if (frameno < 0)
						{
							if (stop_at_end)
							{
								status = STOPPED;
								frameno = 0;
							}
							else
								frameno = nfiles-1;
						}
					}
	
					SetSB(frameno, nfiles, 1);

					if (bench)		
					{
						if (count > nframes)
						{
							times = TimerEnd(timer2);
							printf("%d frames in %f sec (%f fps) (%f sec)\n", 
								count, times, (double)count/times, times/(double)count);
							printf("%d frames dropped\n", dropped);
							TimerStart(&timer2);
							count = 0;
						}
						else
						// first time around 
						if (count == -1)
						{
							count = 0;
							TimerStart(&timer2);
						}	
					}
				}
            break;
		}
	}
	exit(1);
}

/*==============================================*/
int main(int argc, char **argv)
{
	XEvent report;
	int count, nframes, dropped, sleep_ret;
	int CompletionType, bsz, keyhit, frameno;
	int bench = 0;
	my_timer timer1, timer2;
	double times;

	unsigned char **image_list, *p1, *p2;
	char **file_list;
	int nfiles, a, b, c, chunk;
	my_timer clocks_for_a_frame;
	
	int pnum, isodd;
	char *pattern;

	int on = 1;
	TimerInit();

	file_list = NULL;

	gargc = argc;
	gargv = argv;

	if (argc == 1)
	{
		usage(argv);
		exit(1);
	}

	lut_init(&load_lut);

	/* if your file loader uses options, you'll have to add them in 
	 * here, but with a null case (see x and y) */
	while((pnum = getopt(argc, argv, OPT_VALUES)) != -1)
	{
		switch(pnum)
		{
			case 'a':
				if (atoi(optarg))
				{
					auto_check_stale = 1;
					auto_reload_stale = 1;
				}
				else
				{
					auto_check_stale = 1;
					auto_reload_stale = 0;
				}
				break;

			case 'g':
				lut_gamma(&load_lut, atof(optarg));
				break;

			case 's':
				start_bound = atoi(optarg);
				break;

			case 'e':
				end_bound = atoi(optarg);
				break;

			case 'f':
				frames_a_second = atof(optarg);
				time_for_a_frame = 1./frames_a_second;
				break;

			case 'n':
				frames_a_second = 29.97;
				time_for_a_frame = 1./29.97;
				break;

			case 'p':
				frames_a_second = 25.;
				time_for_a_frame = 1./25.;
				break;

			case 'b':
				bench = 1;
				width = atoi(optarg);
				height = .75*width;
				break;
			
			default:
				if (!index(OPT_VALUES, pnum))
				{
					usage(argv);
					exit(1);
				}
				break;
		}
	}

#ifdef X86_LINUX_TIMER
	clocks_for_a_frame = SecsToClocks(time_for_a_frame);
#else
	clocks_for_a_frame.tv_sec  = time_for_a_frame;
	clocks_for_a_frame.tv_usec = time_for_a_frame * 1e6;
#endif

	
	if (!bench)
	{
//		file_list = find_matching_files("/home/rkarl/inst/local/#.ppm", &nfiles, start_bound, end_bound);
//		file_list = find_matching_files("/home/rkarl/matterial/matterial/src/rgb.#.iff", &nfiles, 
//				start_bound, end_bound);
//		file_list = find_matching_files("/home/rkarl/matterial/matterial/src/rgb.#.rla", &nfiles,
//				start_bound, end_bound);
//		file_list = find_matching_files("/home/rkarl/matterial/matterial/src/rgb.#.tif", &nfiles,
//				start_bound, end_bound);

		if (!file_list)
		{
			if (argc <= optind)
			{
				printf("no pattern specified!\n");
				usage(argv);
				exit(1);
			}
			pattern = argv[optind];
			file_list = find_matching_files(pattern, &nfiles, start_bound, end_bound);
			if (!file_list)
			{
				printf("no files found matching pattern %s!\n", pattern);
				exit(1);
			}	
		}
	}
	else
	{
		nfiles = 15;
		file_list = NULL;
	}

	main_fast(argc, argv, file_list, nfiles, bench);

	status = LOADING;

	image_list = (unsigned char **)malloc(nfiles*sizeof(unsigned char *));
	for (a=0; a<nfiles; a++)
	{
		if (!bench)
		{
			if (!ReadFiles(file_list, a, &width, &height, image_list, argc, argv, &load_lut))
			{
				nfiles = a;
				break;
			}
		}
		else
		{
			image_list[a] = (unsigned char *)malloc(4*width*height);
			for (b=0; b<4*width*height; b++)
				image_list[a][b] = 255*drand48();
		}

		if (!a)
		{
			GoX(argc, argv);
			CompletionType = XShmGetEventBase(display) + ShmCompletion;
		}

		memcpy(img->data, image_list[a], img->bytes_per_line*img->height);
		XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, True);

		XNextEvent(display, &report);
		while (report.type != CompletionType) 
			XNextEvent(display, &report);

		SetSB(a, nfiles, 1);
	}	

	if (!nfiles)
	{
		fprintf(stderr, "no files read from pattern %s\n", pattern);
		 exit(1);
	}	

	/* fill the buffer appears to be BRG + pad byte */
	/* does this change with endianess?             */

	memcpy(img->data, image_list[0], img->bytes_per_line*img->height);

	status = PLAYING;

	dropped = 0;
	nframes = 200;
	count = -1;
	frameno = 0;
	TimerStart(&timer2);
	TimerStart(&timer1);

	SetSB(frameno, nfiles, 1);
	XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, True);

	isodd = height % 2;
	c = img->height / 2;
	chunk = 2*img->bytes_per_line;

	while(1)
	{
		XNextEvent(display,&report);
		switch(report.type) 
		{
			case Expose:
				if (report.xany.window == sb_win)
					SetSB(frameno, nfiles, 1);
				else 
				if (report.xany.window == win)
				{
					/* if we're playing, dont worry about it. we'll fix it next frame 	
					 * otherwise, redisplay this frame but dont generate a new event */	
					if (status != PLAYING)
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, False);
				}
				break;

			case KeyPress:
				keyhit = XKeycodeToKeysym(display, report.xkey.keycode,  0);
				switch(keyhit)
				{
					/* h hides the status bar and info stuff */
					case H_KEY:
						if (!hidden)
						{
							XResizeWindow(display, main_win, width, height);
							whisper.min_width = width;
							whisper.min_height = height;
							whisper.max_width = width;
							whisper.max_height = height;
							XSetStandardProperties(display, main_win, "mFlip", "", None, 
                       				argv, argc, &whisper);
						}
						else
						{
							XResizeWindow(display, main_win, width, height+SCROLL_BAR_HEIGHT+2*text_height);
							whisper.min_width = width;
							whisper.min_height = height+SCROLL_BAR_HEIGHT+2*text_height;
							whisper.max_width = width;
							whisper.max_height = height+SCROLL_BAR_HEIGHT+2*text_height;
							XSetStandardProperties(display, main_win, "mFlip", "", None, 
                       				argv, argc, &whisper);

						}				
						hidden = !hidden;
						break;

					/* space bar starts and stops us */
					case SPACE_BAR:
						if (status == PLAYING)
						{
							status = STOPPED;
						}
						else
						if (status == STOPPED)
						{
							status = PLAYING;
							XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, True);
							TimerStart(&timer2);
							TimerStart(&timer1);
							count = 0;
						}
					break;
		
					/* up takes up  back to the beginning */
					case UP_ARROW:
					case UP_KP:
						if (status != STOPPED) break;
						frameno = 0;
						memcpy(img->data, image_list[frameno], img->bytes_per_line*img->height);
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, False);
						SetSB(frameno, nfiles, 1);
						break;

					/* down goes to the end */
					case DOWN_ARROW:
					case DOWN_KP:
						if (status != STOPPED) break;
						frameno = nfiles-1;
						memcpy(img->data, image_list[frameno], img->bytes_per_line*img->height);
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, False);
						SetSB(frameno, nfiles, 1);
						break;

					/* left goes back a frame */
					case LEFT_ARROW:
					case LEFT_KP:
						if (status != STOPPED) break;
						if (frameno == 0) 
							frameno = nfiles-1;
						else
							frameno--;
						memcpy(img->data, image_list[frameno], img->bytes_per_line*img->height);
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, False);
						SetSB(frameno, nfiles, 1);
						break;

					/* right goes back a frame */
					case RIGHT_ARROW:
					case RIGHT_KP:
						if (status != STOPPED) break;
						frameno = (frameno+1) % nfiles;
						memcpy(img->data, image_list[frameno], img->bytes_per_line*img->height);
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, False);
						SetSB(frameno, nfiles, 1);
						break;

					/* f displays the current frame name on stdout */
					case F_KEY:
						if (status != STOPPED) break;
						fprintf(stdout, "%s\n", file_list[frameno]);
						break;

					/* TOCO: CLEANUP! */
					case ESC:
					case Q_KEY:	
						exit(1);
				}
				break;

			default:
				if (report.type == CompletionType)
				{	
					if (!bench)
					{
						/* this returns the number of frames we dropped, if any */
						sleep_ret = TimerSleepAndReset(&timer1, clocks_for_a_frame);
						if (sleep_ret)
						{
							count += sleep_ret;
							frameno = (frameno + sleep_ret) % nfiles;
							dropped += sleep_ret;
						}
					}
				
					SetSB(frameno, nfiles, 1);

//					memcpy(img->data, image_list[frameno], img->bytes_per_line*img->height);
					
					p1 = img->data;
					p2 = image_list[frameno];
					for (b=0; b<c; b++)
					{					
						memcpy(p1, p2, chunk);
						p1 += chunk;
						p2 += chunk;
					}

					if (isodd)
					{
						p1 -= img->bytes_per_line;
						p2 -= img->bytes_per_line;
						memcpy(p1, p2, img->bytes_per_line);
					}

					if (status == PLAYING)
					{
						XShmPutImage(display, win, gc, img, 0, 0, 0, 0, width, height, True);

						count++;
						//frameno++;
						frameno += play_direction;
						if (frameno >= nfiles) frameno = 0;
					}

					if (bench)		
					{
						if (count > nframes)
						{
							times = TimerEnd(timer2);
							printf("%d frames in %f sec (%f fps) (%f sec)\n", 
								count, times, (double)count/times, times/(double)count);
							printf("%d frames dropped\n", dropped);
							TimerStart(&timer2);
							count = 0;
						}
						else
						// first time around 
						if (count == -1)
						{
							count = 0;
							TimerStart(&timer2);
						}	
					}
				}
            break;
		}
	}

	return 1;
}
