/* platform.c 
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <dirent.h>
#include <regex.h>

char *sort_pattern;

/* yoink the number out of a string, given a pattern with
 * # in it to specify the number..                      */
int str_extract_num(char *str, char *pattern)
{
	char *tmp, *p;
	int val;

	/* would be nicer to do this in place.. but it wont be done much.. */
	tmp = strdup(str);

	p = pattern;
	while ((*p) && (*p != '#'))
	{
		p++;
		tmp++;
	}
	
	p = tmp;
	while ((*p) && (*p >= '0') && (*p <= '9'))
	{
		p++;
	}
	*p = 0;

	val = atoi(tmp);

/* this make irix puke all over the next
   time we try and malloc
	free(tmp);
*/

	return val;
}

/* to keep from depending on glib... */
char * k_strndup(char * str, int n)
{
	char *gnew;

	gnew = (char *)malloc(n+1);
	memset(gnew, 0, n+1);
	strncpy(gnew, str, n);

	return gnew;
}

/* to keep from depending on glib, delimiter is a char */
char ** k_strsplit(char *str, char *delimiter, int *n)
{
	char **gnew, *cpy;
	int howmany = 0;
	int a;

	cpy = strdup(str);

	for (a=0; a<strlen(str); a++)
		if (str[a] == delimiter[0]) 
			howmany++;

	howmany++;
	gnew = (char **)malloc(howmany*sizeof(char *));

	for (a=0; a<howmany-1; a++)
	{
		gnew[a] = strtok(cpy, delimiter);
		cpy += strlen(gnew[a])+1;
	}
	gnew[howmany-1] = cpy;

	*n = howmany;
	return gnew;
}

/* to keep from depending on glib... */
char * k_strjoin(char *divider, char **strs, int n)
{
	int newlen = 0;
	char *gnew;
	int a;

	for (a=0; a<n; a++)
		newlen += strlen(strs[a]);
	newlen += n*strlen(divider);
	newlen++;

	gnew = (char *)malloc(newlen);
	memset(gnew, 0, newlen);

	for (a=0; a<n; a++)
	{
		strcat(gnew, strs[a]);
		if (a != n-1)
			strcat(gnew, divider); 
	}

	return gnew;
}

/* compare function for qsort() */
int my_strcmp(const void * a, const void * b)
{
	char *one, *two;
	int vone, vtwo;
	
	one = *(char * const *)a;
	two = *(char * const *)b;

	vone = str_extract_num(one, sort_pattern);
	vtwo = str_extract_num(two, sort_pattern);

	if (vone == vtwo)
		return 0;
	else
	if (vone < vtwo)
		return -1;
	else
		return 1;

/*
	if (strlen(one) == strlen(two))
		return strcmp(one, two);
	else
	if (strlen(one) < strlen(two))
		return -1;
	else
		return 1;
*/

//	return strcmp(*(char * const *)a, *(char * const *)b);
}
 
/* convert a pattern string into a form that we can run through a regex */
char * make_regexable_str(char *pattern)
{
	char *tmp, *pp;
	char regex_bit[] = "[0-9]+";
	char **hold;
	int len, a, n;

	tmp = NULL;
	len = 0;

	pp = pattern;

	/* change all the #'s to regex_bit[]'s */
	while (1)
	{
		len = 0;
		while (( *(pp + len) != '#') && (*(pp + len) != 0)) len++;
		if (*(pp+len) == 0) break; 
		if (tmp == NULL)
		{
			tmp = (char *)malloc((len+1+strlen(regex_bit))*sizeof(char));
			memset(tmp, 0, len+1+strlen(regex_bit));
		}
		else
			tmp = (char *)realloc(tmp, (strlen(tmp)+len+1+strlen(regex_bit))*sizeof(char));
		strncat(tmp, pp, len);
		strcat(tmp, regex_bit);
		pp += len;
		while (*pp == '#') pp++;
	} 

	tmp = (char *)realloc(tmp, (strlen(tmp)+strlen(pp))*sizeof(char));
	strcat(tmp, pp);

	/* escape all the .;s to be \\. */
	hold = k_strsplit(tmp, ".", &n);
	free(tmp);
	tmp = k_strjoin("\\.", hold, n);

	/* add anchor to the front and back */
	len = strlen(tmp);
	tmp = (char *)realloc(tmp, len+3);
	tmp[len+2] = 0;
	tmp[len+1] = '$';
	for (a=len; a>0; a--)
		tmp[a] = tmp[a-1];
	tmp[0] = '^';

	return tmp;
}

/* this guy searches the file system for files mattching 'pattern'. 
 * pattern is a full path. 
 * # expands to [0-9]+ in regex-speak.. so, you could say something like
 *    pattern = "/home/rkarl/pron/horse.#.rla
 *    or
 *    pattern = "/home/rkarl/morepron/goat.iff.#
 *    and so on.. 
 *    returns an array of howmany strings, sorted. 
 *
 * lo is used to specify that we will not match files where # < lo.
 * hi is used to specify that we will not match files where # > hi.	
 */
char ** find_matching_files(char *pattern, int *howmany, int lo, int hi)
{
	char *path, *filename, *testable;
	char **hits = NULL;
	int len, val;
	regex_t reggie;
	struct dirent *dire;
	DIR *did;

	char Directory_Sep_In_My_OS = '/';

	/* if there arent any #'s to match, return pattern */
	if (strstr(pattern, "#") == NULL)
	{
		*howmany = 1;
		hits = (char **)malloc(sizeof(char *));
		hits[0] = strdup(pattern);
		return hits;
	}

	/* find what is the path, and what is the filepattern */
	len = strlen(pattern)-1;
	while ((len >= 0) && (pattern[len] != Directory_Sep_In_My_OS)) len--;	

	if (len == -1)
	{
		path = strdup(".");
		filename = strdup(pattern);
		/* the sort_pattern also needs a proper path prefix */
		sort_pattern = (char*) malloc(strlen(pattern)+2+1);
		strcpy(sort_pattern, "./");
		strcat(sort_pattern, pattern); 
	}
	else	
	{
		path = k_strndup(pattern, len);
		filename = strdup(pattern+len+1);
		sort_pattern = strdup(pattern); /* needed for bounds AND qsort */
	}
	/* setup our regular expression */
	testable = make_regexable_str(filename);

	regcomp(&reggie, testable, REG_NOSUB | REG_EXTENDED);

	/* open the directory... */
	*howmany = 0;
	did = opendir(path);
	if (did == NULL)
	{
		printf("cannot open dir %s\n", path);
		*howmany = 0;
		return NULL;
	}

	/* and check out all the files in it */
	dire = readdir(did);
	while (dire != NULL)
	{
		if (!regexec(&reggie, dire->d_name, 0, NULL, 0))
		{

			/* check that we're within the bounds */
			val = str_extract_num(dire->d_name, filename);

			if ((val < lo) || (val > hi)) 
			{
				dire = readdir(did);
				continue;
			}

			if (hits == NULL)
			{
				hits = (char **)malloc(sizeof(char *));
			}
			else
			{
				hits = (char **)realloc(hits, ((*howmany)+1)*sizeof(char *));
			}

//			hits[(*howmany)] = strdup(dire->d_name);

			hits[(*howmany)] = (char *)malloc(strlen(path)+strlen(dire->d_name)+2);
			memset(hits[(*howmany)], 0, strlen(path)+strlen(dire->d_name)+2);
			strcat(hits[(*howmany)], path);
			hits[(*howmany)][strlen(hits[(*howmany)])] = Directory_Sep_In_My_OS;
			strcat(hits[(*howmany)], dire->d_name);

			(*howmany)++;
		}
		dire = readdir(did);
	}

	closedir(did);

/* this needs to be, sometimes.. i dont know  why
	sort_pattern = (char *)malloc(strlen(path)+strlen(pattern)+2);
	memset(sort_pattern, 0, strlen(path)+strlen(pattern)+2);
	strcat(sort_pattern, path);
	sort_pattern[strlen(sort_pattern)] = Directory_Sep_In_My_OS;
	strcat(sort_pattern, pattern);
*/ 

	/* now, sort the list of hits so we get frames in ascending order */
	qsort(hits, (*howmany), sizeof(*hits), my_strcmp);

	free(sort_pattern);

	free(filename);
	free(testable);
	free(path);
	regfree(&reggie);

	return hits;
}


