/* read_mayaiff.c 
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* 
 * this should not handle standard iff's at this point. maybe later, if there is 
 * a need [ie lots of lightwave/amiga ppl start d/ling.
 *
 * does not handle the 16-bit variant that maya can spit out. 
 *
 * also doenst handle uncompressed files [why would you use them?]
 * */

#include <glib.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "lut.h"

#define FOR4 ('F' << 24) | ('O' << 16) | ('R' << 8) | '4'
#define CIMG ('C' << 24) | ('I' << 16) | ('M' << 8) | 'G'
#define TBMP ('T' << 24) | ('B' << 16) | ('M' << 8) | 'P'
#define TBHD ('T' << 24) | ('B' << 16) | ('H' << 8) | 'D'
#define RGBA ('R' << 24) | ('G' << 16) | ('B' << 8) | 'A'
#define ZBUF ('Z' << 24) | ('B' << 16) | ('U' << 8) | 'F'
#define CLPZ ('C' << 24) | ('L' << 16) | ('P' << 8) | 'Z'

/*===========================================*/
void UsageMIFF()
{
}

/* the iff is broken down into tiles, each with a starting coordinate
 * and a size (tstart and tsize). len is the compressed lenght of the
 * tile. this guy decodes a RLE z-buffer tile. the zvalues are floats,
 * but are encoded one byte at a time, so we shove them into ztmp and
 * reassemble after we have the entire image. */

/* --- function snipped --- */

/* as above, we decode a RLE tile, but this time it consists of RGBA data, stored
 * in ABGR order. if there is no alpha values, we just skip the deciding of the
 * first channel. 
 * channels[] = {achan, bchan, gchan, rchan }                    */
void UnRLETileData(FILE *fid, gint len, gint *tstart, gint *tsize,
			gint w, gint h, unsigned char *buf, int no_alpha, lut_t *lut)
{
	gint a, accum;
	gint size;
	gint x, y;
	gint channel = 0;
	guchar c, cnt, next;

//	if (no_alpha) channel = 1;

	accum = 0;
	size = tsize[0]*tsize[1];

	while (len > 0)
	{
		fread(&c, 1, 1, fid);
		len--;
		cnt = (c & 127) + 1;

		/* the first bit indicates if we have a repeat (1) or a */
		/* copy over (0)                                        */
		if (c & 128)
		{
			fread(&next, 1, 1, fid);
			len--;   
			for (a=0; a<cnt; a++)
			{
				y = (h-1) - (tstart[1] + accum/tsize[0]);
				x = tstart[0] + accum%tsize[0];
		//		((guchar *)(mo->channel[channels[channel]]))[w*y+x + w*h*framenum] = next;
		
				/* order is as follows: (for LE)
				 * 	 IFF	ximage
				 * 	  blue    alpha   0			
				 * 	  green   blue    1
				 * 	  red     green   2
				 * 	  pad     red     3 
				 *
				 * and for BE: ?? */
				if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
					if (channel)
						buf[(w*y+x)*4+(channel-1)] = lut->lut8[next];	
					else
						buf[(w*y+x)*4+3] = next;	
					
//						buf[(w*y+x)*4+(channel-1)] = lut->lut8[next];	
				else
					if (channel)
						buf[(w*y+x)*4+(channel)] = lut->lut8[next];	
					else
						buf[(w*y+x)*4] = next;	
//					buf[(w*y+x)*4+(channel)] = lut->lut8[next];	

				accum++;
			}
		}
		else
		{
			for (a=0; a<cnt; a++)
			{
				fread(&next, 1, 1, fid);
				len--;   
				y = (h-1) - (tstart[1] + accum/tsize[0]);
				x = tstart[0] + accum%tsize[0];
		//		((guchar *)(mo->channel[channels[channel]]))[w*y+x + w*h*framenum] = next;
				if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
					if (channel)
						buf[(w*y+x)*4+(channel-1)] = lut->lut8[next];	
					else
						buf[(w*y+x)*4+3] = next;	

//					buf[(w*y+x)*4+(channel-1)] = lut->lut8[next];
				else
					if (channel)
						buf[(w*y+x)*4+(channel)] = lut->lut8[next];	
					else
						buf[(w*y+x)*4] = next;	

//					buf[(w*y+x)*4+(channel)] = lut->lut8[next];	

				accum++;
			}
		}

		if (accum == size)
		{
			accum = 0;
			channel++;
		}
	}

}



/* we are passed the filename of a file which is known to be
 * readable. we are to look at it to see if we recognize the file.
 *
 * return 0 if we cannot recognize the file, otherwise we can
 * recognize. the return value will be passed to the actual
 * reader func */
int RecMIFF(char *filename, int argc, char **argv)
{
	FILE *fid;
	gulong tag;

	fid = fopen(filename, "r");

	fread(&tag, 4, 1, fid);
	tag = GULONG_FROM_BE(tag);
	fclose(fid);

	if (tag == (FOR4))
		return 1;
	return 0;
}

/* load data. search the MO props for the filename
 * ml_recognize() function. 
 * offset         -- how many frames into the sequence we start at. we should read
 *                   *_FRAMES_BUFFERED frames that must include offset. ie if 
 *                   offset + FRAMES_BUFFERD > eof, read the last FRAMES_BUFFERED
 * recognize_code -- the return value from recognize() func
 * reload == 1, we're swapping in
 * reload == 0, fresh start, set properties and such 
 * return values? 0 -- sucess, failure otherwise.. */

int ReadMIFF(char *filename, int *w, int *h, unsigned char **buf, 
			int argc, char **argv, lut_t *lut) 
{
	/* can you have a messier declaration section?? */
	FILE *fid;
	gint ntiles, start[2], size[2];
	gushort us, x[2], y[2];
	gulong tag, len, flags, compress;
	gfloat f;
	gint first_for4 = 1;
	struct stat file_stat;

	fid = fopen(filename, "r");
	if (!fid)
	{
		return 0;
	}

	stat(filename, &file_stat);

	/* actual file parsing begins here. MayaIFF is a tiled bitmap, which makes
	 * putting it back together in order a big ol mess. just allocate temp
	 * buffers to store in now, then shove it into channels when we're done.
	 * not the most efficnet, but should work. */
		
	while (!feof(fid))
	{
		fread(&tag, 4, 1, fid);
		tag = GULONG_FROM_BE(tag);

		switch(tag)
		{
			case FOR4:
				fread(&len, 4, 1, fid);

				if (first_for4)
				{
					len = GULONG_FROM_BE(len);
					first_for4 = 0;

					/* if the file size is right, the file is probably ok.. */
					if (file_stat.st_size != len+8)
					{
						fprintf(stderr, "corrupt file %s\n", filename);
						fclose(fid);
						return 0;
					}
				}
				break;
				
			case CIMG:
				break;

			case TBMP:
				break;
			
			case CLPZ:
				fread(&len, 4, 1, fid);
				fread(&f, 4, 1, fid);
				fread(&f, 4, 1, fid);
				break;

			/* the tiled bitmap header. looks something like:
			 * word   -- h
			 * word   -- w
			 * half   -- 1
  			 * half   -- 1
			 * word   -- flags [1 - rgb, 2 - a, 3 - z]
			 * half   -- 0
			 * half   -- ntiles
			 * word   -- compress [1 - yes:rle, 0 - no] ??
			 * word   -- 0
			 * word   -- 0 */
			case TBHD:
				fread(&len, 4, 1, fid);
				fread(&len, 4, 1, fid);
				*w = GULONG_FROM_BE(len);
				fread(&len, 4, 1, fid);
				*h = GULONG_FROM_BE(len);
				fread(&len, 4, 1, fid);
				fread(&len, 4, 1, fid);
				flags = GULONG_FROM_BE(len);
				fread(&us, 2, 1, fid);
				fread(&us, 2, 1, fid);
				ntiles = GUINT16_FROM_BE(us);
				fread(&len, 4, 1, fid);
				compress = GULONG_FROM_BE(len);
				fread(&len, 4, 1, fid);
				fread(&len, 4, 1, fid);

				*buf = (unsigned char *)malloc((*w)*(*h)*4);
				memset(*buf, 0, 4*(*w)*(*h));
				break;

			case RGBA:
				fread(&len, 4, 1, fid);
				if (feof(fid)) break;
				len = GULONG_FROM_BE(len);
				fread(&x[0], 2, 1, fid);
				fread(&y[0], 2, 1, fid);
				fread(&x[1], 2, 1, fid);
				fread(&y[1], 2, 1, fid);
				x[0] = GUINT16_FROM_BE(x[0]);
				x[1] = GUINT16_FROM_BE(x[1]);
				y[0] = GUINT16_FROM_BE(y[0]);
				y[1] = GUINT16_FROM_BE(y[1]);
				len -= 8;

				start[0] = x[0];
				start[1] = y[0];
				size[0] = abs(x[1] + 1 - x[0]);
				size[1] = abs(y[1] + 1 - y[0]);

				UnRLETileData(fid, len, start, size, *w, *h, *buf, !(flags&0x2), lut);
				if (len % 4)
					fseek(fid, 4-(len%4), SEEK_CUR);
				break;

			default:
				fread(&len, 4, 1, fid);
				if (feof(fid)) break;
				len = GULONG_FROM_BE(len);
				if (len % 4)
					fseek(fid, len + 4 - (len%4), SEEK_CUR);
				else
					fseek(fid, len, SEEK_CUR);
				break;
		}
	}

	fclose(fid);

	return 1;
}

