/* read_rla.c
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>

#include "lut.h"

typedef struct st_rla_header
{
	gint16 win_left, win_right, win_bottom, win_top;
	gint16 awin_left, awin_right, awin_bottom, awin_top;
	gint16 frame;
	gint16 storage_type;
	gint16 num_chan;
	gint16 num_matte;
	gint16 num_aux;
	gint16 revision;
	gchar gamma[16];
	gchar red_pri[24];
	gchar green_pri[24];
	gchar blue_pri[24];
	gchar white_pt[24]; 
	glong job_num;
	gchar name[128];
	gchar desc[128];
	gchar program[64];
	gchar machine[32];
	gchar user[32];
	gchar date[20];
	gchar aspect[24];
	gchar aspect_ratio[8];
	gchar chan[32];
	gint16 field;
	gchar time[12];
	gchar filter[32];
	gint16 chan_bits;
	gint16 matte_type;
	gint16 matte_bits;
	gint16 aux_type;
	gint16 aux_bits;
	gchar aux[32];
	gchar space[36];
	glong next;
} rla_header;


/*===============================================================*/
void UsageRLA()
{

}

/*===============================================================*/
/* internal funcs                                                */
/* 0 on sucess, 1 on failure, read and fix order in a header     */
int read_rla_header(FILE *fid, rla_header *head)
{
	int br;

	if (fread(head, 1, sizeof(rla_header), fid) != sizeof(rla_header))
		return 1;

	head->win_left = GINT16_FROM_BE(head->win_left);
	head->win_right = GINT16_FROM_BE(head->win_right);
	head->win_bottom = GINT16_FROM_BE(head->win_bottom);
	head->win_top = GINT16_FROM_BE(head->win_top);
	head->awin_left = GINT16_FROM_BE(head->awin_left);
	head->awin_right = GINT16_FROM_BE(head->awin_right);
	head->awin_bottom = GINT16_FROM_BE(head->awin_bottom);
	head->awin_top = GINT16_FROM_BE(head->awin_top);

	head->frame = GINT16_FROM_BE(head->frame);
	head->storage_type = GINT16_FROM_BE(head->storage_type);

	head->num_chan = GINT16_FROM_BE(head->num_chan);
	head->num_matte = GINT16_FROM_BE(head->num_matte);
	head->num_aux = GINT16_FROM_BE(head->num_aux);
	head->revision = GINT16_FROM_BE(head->revision);

	head->job_num = GLONG_FROM_BE(head->job_num);
	head->field = GINT16_FROM_BE(head->field);

	head->chan_bits = GINT16_FROM_BE(head->chan_bits);
	head->matte_type = GINT16_FROM_BE(head->matte_type);
	head->matte_bits = GINT16_FROM_BE(head->matte_bits);
	head->aux_type = GINT16_FROM_BE(head->aux_type);
	head->aux_bits = GINT16_FROM_BE(head->aux_bits);
	head->next = GLONG_FROM_BE(head->next);

	return 0;
}

/*===============================================================*/
/* functions available to other                                  */

/* we are passed the filename of a file which is known to be
 * readable. we are to look at it to see if we recognize the file.
 *
 * return 0 if we cannot recognize the file */
int RecRLA(char *filename, int argc, char **argv)
{
	FILE *fid;
	rla_header rh;
	int a;
	int empty_sum = 0;

	fid = fopen(filename, "r");
	if (read_rla_header(fid, &rh))
	{
		fclose(fid);
		return 0;
	}
	fclose(fid);

	for (a=0; a<36; a++)
		empty_sum += rh.space[a];

	/* the revision should be 0xfffe, and space should be empty */
	if (((rh.revision & 65535) != 0xfffe) || (empty_sum))
		return 0;

	/* now, for limitiations of my code */
	if (rh.aux_type != 0)
	{
		fprintf(stderr, "sorry, not supporting type %d aux channels at this point\n", 
				rh.aux_type);
		return 0;
	}

	if (((rh.num_matte) && (rh.matte_bits != 8)) || (rh.chan_bits != 8))
	{
		fprintf(stderr, "sorry, not supporting non-8 bit/channel images at this point\n");
		return 0;
	}
  
	if (rh.next)
	{
		fprintf(stderr, "sorry, not supporting multi-frame images at this point\n");
		return 0;
	}

	/* 1 -- rgb
	 * 2 -- rgba 
	 * 3 -- rgbaz, assume that z is the first aux channel */
	
	if (!rh.num_matte)
		return 1;
	else
	if (!rh.num_aux)
		return 2;
	else
		return 3;
}

/* load, partner */
int ReadRLA(char *filename, int *w, int *h, unsigned char **buf,
			int argc, char **argv, lut_t *lut) 
{
	FILE *fid;
	rla_header head;
	gint a, chan;
	gint line, cnt, start;
	signed char length;			// irix does their own thing where 
	guchar data;				// char != signed char.. bastards!
	gint32 *offset_table;
	
	fid = fopen(filename, "r");
	if (!fid)
	{
		fprintf(stderr, "cannot read file %s\n", filename);
		return 0;
	}

	read_rla_header(fid, &head);
	*w = head.awin_right - head.awin_left + 1;
	*h = head.awin_top - head.awin_bottom + 1;
		
	offset_table = (gint32 *)malloc((*h)*sizeof(gint32));
	fread(offset_table, 1, sizeof(gint32)*(*h), fid);
	for (a=0; a<(*h); a++)
		offset_table[a] = GINT32_FROM_BE(offset_table[a]);

	*buf = (unsigned char *)malloc(4*(*w)*(*h));
	memset(*buf, 0, 4*(*w)*(*h));

	for (line=0; line<*h; line++)
	{
		fseek(fid, offset_table[line], SEEK_SET);

		for (chan=0; chan<4; chan++)
		{	
			start = (*h-1-line)*(*w);
			cnt = 0;
			while (cnt <= (*w)-1)
			{
				fread(&length, 1, 1, fid);
				if (length < 0)
				{
					/* next abs(length) bytes are just copied over */
					for (a=0; a<abs(length); a++)	
					{	
						fread(&data, 1, 1, fid);
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
							if (chan != 3)
								(*buf)[(start+cnt)*4 + (2-chan)] = lut->lut8[data];
							else
								(*buf)[(start+cnt)*4 + chan] = data;
						else
							if (chan != 3)
								(*buf)[(start+cnt)*4 + 4-(chan+1)] = lut->lut8[data];
							else
								(*buf)[(start+cnt)*4] = data;
					
							
						cnt++;
					}
				}
				else
				{
					/* repeat next byte length+1 times */
					fread(&data, 1, 1, fid);
					for (a=0; a<length+1; a++)
					{
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
							if (chan != 3)
								(*buf)[(start+cnt)*4 + (2-chan)] = lut->lut8[data];
							else
								(*buf)[(start+cnt)*4 + chan] = data;
						else
							if (chan != 3)
								(*buf)[(start+cnt)*4 + 4-(chan+1)] = lut->lut8[data];
							else
								(*buf)[(start+cnt)*4] = data;
						cnt++;	
					}
				}
			}	
		}
	}

	fclose(fid);

	return 1;
}
