/* read_tiff.c
 * Copyright (C) 2001 by The matterial Developers 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef HAVE_CONFIG_H
#include "config.h"

#ifdef HAVE_LIBTIFF

#include <tiffio.h>

#include "lut.h"

typedef struct st_ifd_ent
{
	guint16 id;
	guint16 type;
	guint32 length;
	guint32 offset;
} ifd_ent;

/*==========================================================*/
void UsageTIFF()
{

}

/*==========================================================*/
/* we are passed the filename of a file which is known to be
 * readable. we are to look at it to see if we recognize the file.
 *
 * return 0 if we cannot recognize the file */
int RecTIFF(char *filename, int argc, char **argv)
{
	TIFF *fid;
	gchar msg[1024];
	gint16 bpp;

	fid = TIFFOpen(filename, "r");
	
	if (fid)
	{
		if (!TIFFRGBAImageOK(fid, msg))
		{
			TIFFClose(fid);
			return 0;
		}


		TIFFGetField(fid, TIFFTAG_BITSPERSAMPLE, &bpp);
		bpp /= 8;
		if ((bpp != 1) && (bpp != 2) && (bpp != 4))
		{
			TIFFClose(fid);
			return 0;
		}

/*
		if (TIFFIsTiled(fid))
		{
			fprintf(stderr, "%s is a tiled tiff, yell at karl to support it\n",
				filename);
			TIFFClose(fid);
			return 0;
		}
*/

		TIFFClose(fid);
		return 1;
	}

	return 0;
}

/* change my pitch up.. smack my tiff up.. */

/*======================================================================*/
int ReadTIFFStrip(char *filename, int *w, int *h, unsigned char **ibuf, lut_t *lut) 
{
	gint a, b, pos;
	TIFF *fid;
	gchar c;
	gint read_amt;
	gint recognize_code, filename_list_length;
	gint rchan, gchan, bchan, achan;
	gint real_offset, end_point, framenum;
	guint16 planes, bpp;
	guint32 *buf;


	fid = TIFFOpen(filename, "r");
	if (!fid)
	{
		fprintf(stderr, "cannot open %s\n", filename);
		exit(1);
	}

	TIFFGetField(fid, TIFFTAG_IMAGEWIDTH, w);
	TIFFGetField(fid, TIFFTAG_IMAGELENGTH, h);
	TIFFGetField(fid, TIFFTAG_SAMPLESPERPIXEL, &planes);
	TIFFGetField(fid, TIFFTAG_BITSPERSAMPLE, &bpp);

	*ibuf = (unsigned char *)malloc(4*(*w)*(*h));
	memset(*ibuf, 0, 4*(*w)*(*h));

	buf =  (guint32 *) _TIFFmalloc(TIFFStripSize(fid));
	pos = 0;

	for (a=0; a<TIFFNumberOfStrips(fid); a++)
	{
/*
		printf("(%d/%d): %d\n", a, TIFFNumberOfStrips(fid), 
                                      (int)(TIFFReadEncodedStrip(fid, a, buf, (tsize_t)-1)));
*/
		read_amt = TIFFReadEncodedStrip(fid, a, buf, (tsize_t)-1);
//		for (b=0; b<TIFFStripSize(fid)/(bpp/8); b+=planes)
		for (b=0; b<read_amt/(bpp/8); b+=planes)
		{
			switch (bpp/8)
			{
				case 1:      /* <= 8 bit/chan */
					if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
					{
						(*ibuf)[4*pos] =   lut->lut8[((guchar *)buf)[b+2]]; 
 						(*ibuf)[4*pos+1] = lut->lut8[((guchar *)buf)[b+1]]; 
						(*ibuf)[4*pos+2] = lut->lut8[((guchar *)buf)[b]];
						if (planes > 3)
							(*ibuf)[4*pos+3] = ((guchar *)buf)[b+3]; 
						else
							(*ibuf)[4*pos+3] = 255; 
					}
					else
					{
 						(*ibuf)[4*pos+1] = lut->lut8[((guchar *)buf)[b+2]]; // blue
						(*ibuf)[4*pos+2] = lut->lut8[((guchar *)buf)[b+1]]; // green
						(*ibuf)[4*pos+3] = lut->lut8[((guchar *)buf)[b]];   // red
						if (planes > 3)
							(*ibuf)[4*pos] = ((guchar *)buf)[b+3];
						else
							(*ibuf)[4*pos] = 255;
					}
					break;


				case 2:      /* 8 < x <= 16 bit/chan */
					if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
					{
						(*ibuf)[4*pos] =   lut->lut16[((guint16 *)buf)[b+2]]   >> 8; 
	 					(*ibuf)[4*pos+1] = lut->lut16[((guint16 *)buf)[b+1]] >> 8; 
						(*ibuf)[4*pos+2] = lut->lut16[((guint16 *)buf)[b]] >> 8;
						if (planes > 3)
							(*ibuf)[4*pos+3] = ((guint16 *)buf)[b+3] >> 8;
						else
							(*ibuf)[4*pos+3] = 255;
					}
					else
					{
	 					(*ibuf)[4*pos+1] = lut->lut16[((guint16 *)buf)[b+2]] >> 8;// blue
						(*ibuf)[4*pos+2] = lut->lut16[((guint16 *)buf)[b+1]] >> 8; // green
						(*ibuf)[4*pos+3] = lut->lut16[((guint16 *)buf)[b]]   >> 8;   // red
						if (planes > 3)
							(*ibuf)[4*pos] = ((guint16 *)buf)[b+3]   >> 8;  
						else
							(*ibuf)[4*pos] = 255;
					}
					break;

				case 4:      /* 16 < x <= 32 bit/chan ... you're too
							  * cool if you work @ this depth.. untested code.
							  * note the lut hack */
					if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
					{
						(*ibuf)[4*pos] =   lut->lut8[((guint32 *)buf)[b+2]   >> 24]; 
 						(*ibuf)[4*pos+1] = lut->lut8[((guint32 *)buf)[b+1] >> 24]; 
						(*ibuf)[4*pos+2] = lut->lut8[((guint32 *)buf)[b] >> 24];
						if (planes > 3)
							(*ibuf)[4*pos+3] = ((guint32 *)buf)[b+3] >> 24;
						else
							(*ibuf)[4*pos+3] = 255;
					}
					else
					{
	 					(*ibuf)[4*pos+1] = lut->lut8[((guint32 *)buf)[b+2] >> 24];// blue
						(*ibuf)[4*pos+2] = lut->lut8[((guint32 *)buf)[b+1] >> 24]; // green
						(*ibuf)[4*pos+3] = lut->lut8[((guint32 *)buf)[b]   >> 24]; // red
						if (planes > 3)
							(*ibuf)[4*pos] = ((guint32 *)buf)[b+3]   >> 24;
						else
							(*ibuf)[4*pos] = 255; 

					}

					break;
			}
			pos++;
		}
	}

	TIFFClose(fid);
	free(buf);

	return 1;
}

/*======================================================================*/
int ReadTIFFTile(char *filename, int *w, int *h, unsigned char **ibuf, lut_t *lut) 
{
	gint a, b, pos;
	TIFF *fid;
	gchar c;
	gint read_amt;
	gint recognize_code, filename_list_length;
	gint rchan, gchan, bchan, achan;
	gint real_offset, end_point, framenum;
	gint tilew, tileh, tilenum, tx, ty;
	gint ntx, nty;
	guint16 planes, bpp;
	guint32 *buf;

	fid = TIFFOpen(filename, "r");
	if (!fid)
	{
		fprintf(stderr, "cannot open %s\n", filename);
		exit(1);
	}
			
	TIFFGetField(fid, TIFFTAG_IMAGEWIDTH, w);
	TIFFGetField(fid, TIFFTAG_IMAGELENGTH, h);
	TIFFGetField(fid, TIFFTAG_SAMPLESPERPIXEL, &planes);
	TIFFGetField(fid, TIFFTAG_BITSPERSAMPLE, &bpp);
	TIFFGetField(fid, TIFFTAG_TILEWIDTH, &tilew);
	TIFFGetField(fid, TIFFTAG_TILELENGTH, &tileh);

	*ibuf = (unsigned char *)malloc(4*(*w)*(*h));
	memset(*ibuf, 0, 4*(*w)*(*h));

	buf = _TIFFmalloc(TIFFTileSize(fid));

	ntx = ceil((float)(*w)/(float)tilew);
	nty = ceil((float)(*h)/(float)tileh);

	for (ty=0; ty<nty; ty++)
		for (tx=0; tx<ntx; tx++)
		{
			TIFFReadEncodedTile(fid, ty*ntx+tx, buf, (tsize_t)-1);

			for (a=0; a<tileh; a++)
				for (b=0; b<tilew; b++)
				{
					if ((tx*tilew)+b >= (*w)) continue;
					if ((ty*tileh)+a >= (*h)) continue;

					pos = ((ty*tileh)+a)*(*w)+(tx*tilew)+b;
					switch (bpp/8)
					{
						case 1:      /* <= 8 bit/chan */
							if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
							{
								(*ibuf)[4*pos] =   lut->lut8[((guchar *)buf)[planes*(a*tilew+b)+2]]; 
 								(*ibuf)[4*pos+1] = lut->lut8[((guchar *)buf)[planes*(a*tilew+b)+1]]; 
								(*ibuf)[4*pos+2] = lut->lut8[((guchar *)buf)[planes*(a*tilew+b)]];
								if (planes > 3)
									(*ibuf)[4*pos+3] = ((guchar *)buf)[planes*(a*tilew+b)+3];
								else
									(*ibuf)[4*pos+3] = 255;
							}
							else
							{
								(*ibuf)[4*pos+1] = lut->lut8[((guchar *)buf)[planes*(a*tilew+b)+2]]; 
 								(*ibuf)[4*pos+2] = lut->lut8[((guchar *)buf)[planes*(a*tilew+b)+1]]; 
								(*ibuf)[4*pos+3] = lut->lut8[((guchar *)buf)[planes*(a*tilew+b)]];
								if (planes > 3)
									(*ibuf)[4*pos] = ((guchar *)buf)[planes*(a*tilew+b)+3];
								else
									(*ibuf)[4*pos] = 255;
							}
						break;

					case 2:      /* 8 < x <= 16 bit/chan */
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
						{
							(*ibuf)[4*pos] =   lut->lut16[((guint16 *)buf)[planes*(a*tilew+b)+2]] >> 8; 
							(*ibuf)[4*pos+1] = lut->lut16[((guint16 *)buf)[planes*(a*tilew+b)+1]] >> 8; 
							(*ibuf)[4*pos+2] = lut->lut16[((guint16 *)buf)[planes*(a*tilew+b)]]   >> 8;
							if (planes > 3)
								(*ibuf)[4*pos+3] = ((guint16 *)buf)[planes*(a*tilew+b)+3]   >> 8;
							else
								(*ibuf)[4*pos+3] = 255;
						}
						else
						{
							(*ibuf)[4*pos+1] = lut->lut16[((guint16 *)buf)[planes*(a*tilew+b)+2]] >> 8; 
							(*ibuf)[4*pos+2] = lut->lut16[((guint16 *)buf)[planes*(a*tilew+b)+1]]   >> 8; 
							(*ibuf)[4*pos+3] = lut->lut16[((guint16 *)buf)[planes*(a*tilew+b)]]     >> 8;
							if (planes > 3)
								(*ibuf)[4*pos] = ((guint16 *)buf)[planes*(a*tilew+b)+3]     >> 8;
							else
								(*ibuf)[4*pos] = 255;
						}
						break;

					case 4:      /* 16 < x <= 32 bit/chan ... you're too
								  * cool if you work @ this depth.. untested code.
								  * note the lut hack */
						if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
						{
							(*ibuf)[4*pos] =   lut->lut8[((guint32 *)buf)[planes*(a*tilew+b)+2] >> 24]; 
							(*ibuf)[4*pos+1] = lut->lut8[((guint32 *)buf)[planes*(a*tilew+b)+1] >> 24]; 
							(*ibuf)[4*pos+2] = lut->lut8[((guint32 *)buf)[planes*(a*tilew+b)]   >> 24];
							if (planes > 3)
								(*ibuf)[4*pos+3] = ((guint32 *)buf)[planes*(a*tilew+b)+3]   >> 24;
							else
								(*ibuf)[4*pos+3] = 255;
						}
						else
						{
							(*ibuf)[4*pos+1] = lut->lut8[  ((guint32 *)buf)[planes*(a*tilew+b)+2] >> 24]; 
							(*ibuf)[4*pos+2] = lut->lut8[((guint32 *)buf)[planes*(a*tilew+b)+1]   >> 24]; 
							(*ibuf)[4*pos+3] = lut->lut8[((guint32 *)buf)[planes*(a*tilew+b)]     >> 24];
							if (planes > 3)
								(*ibuf)[4*pos] = ((guint32 *)buf)[planes*(a*tilew+b)+3]     >> 24;
							else
								(*ibuf)[4*pos] = 255;

						}
						break;
					}
				}
		}

	_TIFFfree(buf);
	TIFFClose(fid);

	return 1;
}

/*======================================================================*/
int ReadTIFF(char *filename, int *w, int *h, unsigned char **ibuf, int argc, 
			char **argv,  lut_t *lut) 
{
	TIFF *fid;
	int tiled;

	fid = TIFFOpen(filename, "r");
	if (!fid)
	{
		fprintf(stderr, "cannot open %s\n", filename);
		exit(1);
	}

	tiled = TIFFIsTiled(fid);
	TIFFClose(fid);

	if (!tiled)
		return ReadTIFFStrip(filename, w, h, ibuf, lut);

	return ReadTIFFTile(filename, w, h, ibuf, lut);
}

#endif
#endif
